use crate::config::node::NodeType;
use log::{debug, error, info, trace};
use std::{env, io};
use connection::client::Client;


use domo_proto::packet::identifier::Identifier;
use std::cell::RefCell;
use tokio::io::AsyncWriteExt;
use tokio::net::TcpSocket;
use domo_proto::commands::node_management::NodeManagement;
use domo_proto::packet::ToPacket;
use crate::connection::from_tcp_stream;

pub mod config;
pub mod connection;

thread_local! {
    static CONFIG: RefCell<config::Config> = RefCell::new(config::Config::from_path(
        env::var("DOMO_CONFIG_PATH")
            .unwrap_or("config.toml".into())
    ).unwrap_or({ info!("Using default config..."); config::Config::default() }))
}

fn main() {
    setup_logging().expect("could not setup logging");

    debug!("Building tokio runtime...");

    let runtime = tokio::runtime::Builder::new_multi_thread()
        .enable_all()
        .build()
        .unwrap();

    debug!("Built tokio runtime with all features...");

    info!(
        "Staring as {} node!",
        CONFIG.with_borrow(|c| c.node.node_type.clone())
    );


    match CONFIG.with_borrow(|c| c.node.node_type.clone()) {
        NodeType::Master { bind, identifier } => runtime.block_on(async {
            let identifier = Identifier::from(hex::decode(identifier).unwrap().as_slice());
            trace!("identifier: {identifier}");
            let mut sock = TcpSocket::new_v4().unwrap();
            sock.bind(bind.parse().unwrap()).unwrap();

            let mut listener = sock.listen(1024).unwrap();
            info!("Listening at {}", bind);
            loop {
                match listener.accept().await {
                    Ok((mut stream, addr)) => {
                        tokio::spawn(async move {
                            loop {
                                let packet = from_tcp_stream(&mut stream).await.unwrap();
                                stream.write(NodeManagement::RegisterNode {
                                    device_id: Identifier::from(packet.data.data.clone())
                                }.to_packet(
                                    identifier,
                                    Identifier::from(packet.data.data),
                                    Identifier::random(),
                                    Identifier::default()
                                ).build_full_packet().as_slice()).await.unwrap();
                            }
                        });
                    },
                    Err(e) => error!("Could not accept client: {e}")
                }
            }
        }),
        NodeType::Relay { master_address } => {
            let client = Client::new(master_address.parse().unwrap());
            runtime.block_on(async {
                let conn = client.client_connection(
                    CONFIG.with_borrow(|c| c.node.device_id.clone())
                        .map(|s| Identifier::from(hex::decode(s)
                            .expect("Could not decode device_id hex")
                            .as_slice()))
                ).await.unwrap();
                info!("Registered as: {}", conn.device_id());
            })
        },
        NodeType::Subnet { master_address: _ } => {
            runtime.block_on(async {})
        }
    }
}

fn setup_logging() -> Result<(), fern::InitError> {
    fern::Dispatch::new()
        .format(|out, message, record| match record.metadata().target() {
            _ => out.finish(format_args!(
                "[{} - {}] <{}> {}: {}",
                humantime::format_rfc3339(std::time::SystemTime::now()),
                record.level(),
                record.target(),
                record.file().unwrap_or("??"),
                message
            )),
        })
        .chain({
            fern::Dispatch::new()
                .level(CONFIG.with_borrow(|c| c.log.stdout_level.clone()).into())
                .chain(io::stdout())
        })
        .chain({
            fern::Dispatch::new()
                .level(CONFIG.with_borrow(|c| c.log.stdout_level.clone()).into())
                .chain(fern::log_file("./node.log")?)
        })
        .apply()?;

    Ok(())
}
